import shelve
import Queue
import os
import hashlib
import time
from collections import deque, defaultdict
from itertools import imap
from operator import itemgetter

from twisted.python import log
from twisted.internet import defer, threads, reactor
from twisted.plugin import getPlugins

from sdfs.imetadata import IMetadata
from sdfs.enum import DatabaseType
from sdfs.utils import *


COMMIT_COUNTER = 10000 # listing entries before adding to work queue
BUCKET_SIZE = 1 # bytes,

def generate_buckets(bucket_depth=BUCKET_SIZE):
    for i in [chr(x) for x in xrange(256)]:
        if bucket_depth > 1:
            for j in generate_buckets(bucket_depth-1):
                yield i + j
        else:
            yield i

class FileSystem(object):
    def __init__(self, paths, dbfile):
        self.db = shelve.open(dbfile)
        self.paths = paths
        self.fsh = FileSystemHandler(self.db, paths)
    
    def close(self):
        self.db.close()
        return self.fsh.add_to_queue('die', priority=0)
    
    def list_dir(self, paths, depth=0):
        if type(paths) is not list:
            paths = [paths]
        paths = cleanup_path(paths)

        dirlist = {}
        nested_paths = []
        
        for path in paths:
            key = keyify(DatabaseType.FILELIST, path)
            
            if key in self.db:
                dirlist[path] = []
                for key in self.db[keyify(DatabaseType.FILELIST, path)]:
                    item_type = key[0]
                    metadata = self.db[key]
                    
                    if item_type == DatabaseType.DIRECTORY and depth:
                        nested_paths.append(metadata['virtual_path'])
                    
                    elif item_type == DatabaseType.FILE:
                        del metadata['actual_path']
                    
                    metadata['name'] = metadata['virtual_path'].split('/')[-1]
                    del metadata['virtual_path']
                    metadata['type'] = item_type
                    
                    dirlist[path].append(metadata)
            else:
                dirlist[path] = None
            
        if depth:
            dirlist.update(self.list_dir(nested_paths, depth-1))
        return dirlist
    
    def get_metadata(self, path):
        path = cleanup_path(path)
        
        for item_type in [DatabaseType.FILE, DatabaseType.DIRECTORY]:
            key = keyify(item_type, path)
            if key in self.db:
                d = self.db[key]
                if item_type == DatabaseType.FILE:
                    del d['actual_path']
                d['name'] = d['virtual_path'].split('/')[-1]
                return d
        
        return None
    
    def get_file(self, path):
        d = self.db.get(keyify(DatabaseType.FILE, cleanup_path(path)), None)
        return d and d['actual_path'] or d
    
    def rescan(self):
        return self.fsh.complete_rescan()


# run main function in thread, handles all db related
class FileSystemHandler(object):
    def __init__(self, db, paths):
        self.paths = paths
        self.db = db
        self.queue = Queue.PriorityQueue()
        
        threads.deferToThread(self.thread_handler)
    
    def add_to_queue(self, f, *args, **kwargs):
        priority = kwargs.pop('priority', 5)
        d = defer.Deferred()
        self.queue.put((priority, d, f, args, kwargs))
        return d
    
    def thread_handler(self):
        """
        Handles all the jobs thrown into the queue.
        """
        while True:
            priority, d, f, args, kwargs = self.queue.get()
            
            if f == 'die':
                if d:
                    reactor.callFromThread(d.callback, result)
                return
            
            try:
                result = f(*args, **kwargs)
            except:
                if d:
                    reactor.callFromThread(d.errback)
            else:
                if d:
                    reactor.callFromThread(d.callback, result)
            
            
    
    @defer.inlineCallbacks
    def complete_rescan(self):
        log.msg("Starting complete rescan")
        def rescan_setup(): # execute in threadhandler
            log.msg("Creating buckets in database")
            initial = False
            for i in generate_buckets():
                key = '%s%s' % (DatabaseType.BUCKET, i)
                self.db['%sn' % key] = set()
                if key not in self.db: # initial scan
                    self.db[key] = set()
                    initial = True
            
            if initial:
                log.msg('Initial scan, creating root')
                self.db[keyify(DatabaseType.DIRECTORY, '')] = {'virtual_path': '', 'name': ''}
                self.db[keyify(DatabaseType.FILELIST, '')] = set()
        
        yield self.add_to_queue(rescan_setup)
        
        log.msg("Preparing to run scan for each path")
        defer_list = []
        for path, prefix in self.paths:
            defer_list.append(threads.deferToThread(self.optimized_walk, path, prefix))
        
        defer_list = yield defer.DeferredList(defer_list)
        yield defer.DeferredList([d for defers in defer_list for d in defers[1]])
        log.msg("Done scanning paths, time to handle plugins interested in new data.")
        
        def handle_plugins_for_new_files():
            import sdfs.plugins
            plugins = list(getPlugins(IMetadata, sdfs.plugins))
            
            for p in plugins:
                setattr(p, 'db', self.db)
            
            for i in generate_buckets():
                for new_hash in self.db['%s%sn' % (DatabaseType.BUCKET, i)] - self.db['%s%s' % (DatabaseType.BUCKET, i)]:
                    if new_hash[0] != DatabaseType.FILE:
                        continue
                    
                    metadata = self.db[new_hash]
                    r, f = os.path.split(metadata['actual_path'])
                    vr = '/'.join(metadata['virtual_path'].split('/')[:-1])
                    
                    for p in plugins:
                        if p.filetypes.match(f):
                            p.found_file(r, vr, f)
        
        yield self.add_to_queue(handle_plugins_for_new_files)
        
        log.msg("Done handling plugins, cleaning up dead files")
        
        def cleanup_files():
            missing_files_found = 0
            for i in generate_buckets():
                for dead_hash in self.db['%s%s' % (DatabaseType.BUCKET, i)] - self.db['%s%sn' % (DatabaseType.BUCKET, i)]:
                    vr = self.db[dead_hash]['virtual_path']
                    parent_vr = '/'.join(vr.split('/')[:-1])
                    parent_hash = keyify(DatabaseType.FILELIST, parent_vr)
                    
                    del self.db[dead_hash]
                    if parent_hash in self.db:
                        self.db[parent_hash] -= set([dead_hash])
                    
                    missing_files_found += 1
                    
                self.db['%s%s' % (DatabaseType.BUCKET, i)] = self.db['%s%sn' % (DatabaseType.BUCKET, i)]
                del self.db['%s%sn' % (DatabaseType.BUCKET, i)]
            return missing_files_found
        
        missing_files_found = yield self.add_to_queue(cleanup_files)
        
        yield self.add_to_queue(self.db.sync)
        
        log.msg("Done cleaning up, removed %s files." % missing_files_found)
    
    def get_time(self):
        return int(time.time())
    
    def optimized_walk(self, path, prefix):
        log.msg("Starting to scan '%s' with prefix '%s'" % (path, prefix))
        
        defer_list = []
        list_queue = []
        queue_size = 0
        
        def handle_items(path, prefix, items):
            add_time = self.get_time()
            keys_touched = set()
            hashes_found = []
            keys_to_dir_mapping = defaultdict(set)
            
            # make sure all prefix subfolders exist
            tmp_prefix = prefix.strip('/').split('/')
            if tmp_prefix[0] != '':
                while tmp_prefix:
                    vr = '/'.join(tmp_prefix)
                    dir_key = keyify(DatabaseType.DIRECTORY, vr)
                    if dir_key in self.db:
                        break
                    
                    self.db[dir_key] = {
                        'date': add_time,
                        'modified': add_time,
                        'virtual_path': vr
                    }
                    
                    tmp_prefix.pop()
                    parent_vr = '/'.join(tmp_prefix)
                    keys_to_dir_mapping[parent_vr].add(dir_key)
            
            # dump everything into the database
            for r, dirs, files in items:
                # must map folders to parent folder
                vr = r[len(path):]
                vr = '/'.join(vr.split(os.sep))
                vr = ('%s/%s' % (prefix.strip('/'), vr.strip(os.sep)))
                vr = vr.strip('/')
                
                dir_key = keyify(DatabaseType.DIRECTORY, vr)
                hashes_found.append(dir_key)
                
                dir_val = self.db.get(dir_key, {})
                if not dir_val:
                    parent_vr = '/'.join(vr.split('/')[:-1])
                    dir_val = {
                        'date': add_time,
                        'modified': add_time
                    }
                    if vr:
                        keys_to_dir_mapping[parent_vr].add(dir_key)
                
                dir_val['virtual_path'] = vr
                self.db[dir_key] = dir_val
                
                for f in files:
                    if not os.path.isfile(os.path.join(r, f)):
                        continue
                    
                    vrf = '%s/%s' % (vr, f)
                    fr = os.path.join(r, f)
                    file_key = keyify(DatabaseType.FILE, vrf)
                    
                    if file_key not in self.db:
                        self.db[file_key] = {
                            'actual_path': fr,
                            'size': os.path.getsize(fr),
                            'virtual_path': vrf,
                            'date': add_time,
                            'modified': add_time
                        }
                        
                        svr = [x for x in vr.split('/') if x]
                        while svr:
                            keys_touched.add(keyify(DatabaseType.DIRECTORY, '/'.join(svr)))
                            svr.pop()
                    elif self.db[file_key]['actual_path'] != fr:
                        file_val = self.db[file_key]
                        file_val['actual_path'] = fr
                        self.db[file_key] = file_val['actual_path']
                        
                    hashes_found.append(file_key)
                    keys_to_dir_mapping[vr].add(file_key)

            # update modified time
            for key in keys_touched:
                val = self.db[key]
                val['modified'] = add_time
                self.db[key] = val
            
            # update buckets
            buckets = defaultdict(set)
            for h in hashes_found:
                bucket = h[1:1+BUCKET_SIZE]
                buckets[bucket].add(h)
            
            for bucket, keys in buckets.items():
                self.db['%s%sn' % (DatabaseType.BUCKET, bucket)] |= keys

            # add new files to folders
            for vr, file_keys in keys_to_dir_mapping.items():
                dir_key = keyify(DatabaseType.FILELIST, vr)
                file_keys |= self.db.get(dir_key, set())
                self.db[dir_key] = file_keys
        
        for r, dirs, files in os.walk(path):
            list_queue.append((r, dirs, files))
            queue_size += len(dirs) + len(files)
            
            if queue_size >= COMMIT_COUNTER:
                defer_list.append(self.add_to_queue(handle_items, path, prefix, list_queue))
                list_queue = []
                queue_size = 0
        
        if list_queue:
            defer_list.append(self.add_to_queue(handle_items, path, prefix, list_queue))
        
        log.msg("Done scanning '%s'" % path)
        return defer_list