"""
Momentet
"""

import urllib
import glob
from datetime import datetime
from urlparse import urlparse

from twisted.web import xmlrpc
from twisted.python import log
from twisted.plugin import IPlugin, getPlugins
from twisted.internet import threads, defer, reactor

import sdfs.plugins
from sdfs.iurihandler import IURIHandler
from sdfs.fs import FileSystem
from sdfs.utils import uriparse
from sdfs.exceptions import UnknownSchemeException

class SDFSServer(xmlrpc.XMLRPC):
    """General outward facing filesharing API"""
    
    plugins = None
    filesystem = None
    configfile = None
    scanning = False
    last_scan = None
    scan_defer = None
    
    def __init__(self, *args, **kwargs):
        self.configfile = kwargs['configfile']
        del kwargs['configfile']
        xmlrpc.XMLRPC.__init__(self, *args, **kwargs)
        
        paths = []
        i = 1
        while self.configfile.has_option('disks', 'disk%s' % i):
            prefix = self.configfile.get('disks', 'prefix%s' % i)
            for path in glob.glob(self.configfile.get('disks', 'disk%s' % i)):
                paths.append((path, prefix))
            i += 1
        
        reactor.suggestThreadPoolSize(len(paths) + 2)
        self.filesystem = FileSystem(paths, self.configfile.get('general', 'dbfile'))
        self.rescan()
        
        if not self.plugins:
            self.plugins = list(getPlugins(IURIHandler, sdfs.plugins))
        
        self.uninitialize_plugins()
    
    def shutdown(self):
        return self.filesystem.close()
    
    def rescan(self):
        """
        Rescans the whole filesystem, looking from changes compared
        to how the database currently matches the actual filesystem.
        
        Returns a deferred, fired when scan is done.
        """
        
        self.scanning = True
        def done_scanning(ignored):
            self.scanning = False
            self.last_scan = datetime.now()
        
        self.scan_defer = self.filesystem.rescan().addCallback(done_scanning)
    
    def server_status(self):
        """
        Returns current server status which is only affected by
        an ongoing scan.
        """
        if self.scanning:
            status = 'scanning'
        else:
            status = 'idle'
        
        return {
            'date': datetime.now().isoformat(),
            'status': status
        }
    
    def uninitialize_plugins(self):
        for plugin in self.plugins:
            if hasattr(plugin, 'is_initialized'):
                delattr(plugin, 'is_initialized')
    
    def initialize_plugin(self, plugin):
        if not hasattr(plugin, 'is_initialized'):
            setattr(plugin, 'filesystem', self.filesystem)
            setattr(plugin, 'is_initialized', True)
            setattr(plugin, 'configfile', self.configfile)
            setattr(plugin, 'config_section', 'plugin-%s' %
                plugin.scheme)
            if hasattr(plugin, 'dependencies'):
                for dep in plugin.dependencies:
                    self.get_plugin(dep)
            plugin.initialize()


    def get_plugin(self, scheme):
        """
        Takes a protocol scheme and returns the plugin matching it.
        The function ensures that the plugin is initalized.
        """
        
        
        
        for plugin in self.plugins:
            if plugin.scheme == scheme:
                self.initialize_plugin(plugin)
                return plugin
        
        raise UnknownSchemeException()
    
    @defer.inlineCallbacks
    def xmlrpc_get_uri(self, uri):
        """
        Outward facing wrapper for getting an URI
        """
        log.msg("Fetching uri: '%s'" % uri, debug=True)
        
        result = yield self.get_uri(uri)
        
        defer.returnValue((self.server_status(), result))
    
    def get_uri(self, uri):
        """
        Internal function handling uris
        """
        uri = uriparse(uri)
        return threads.deferToThread(self.get_plugin(uri.real_scheme).get_uri, uri)
    
    def xmlrpc_rescan(self):
        """
        Outward facing wrapper for the rescan.
        """
        self.rescan()
