import os
import shutil

from twisted.trial import unittest
from twisted.internet import defer

from sdfs.fs import FileSystem, FileSystemHandler
from sdfs.enum import DatabaseType


class FileSystemScannerTest(unittest.TestCase):
    def _make_file(self, path, size):
        f = open(path, 'w')
        f.write('-'*size)
        f.close()
    
    def setUp(self):
        self.tmpfolder = tmpfolder = os.path.join(os.getcwd(), self.mktemp())
        os.mkdir(tmpfolder)
        
        self.dbfolder = dbfolder = os.path.join(os.getcwd(), self.mktemp())
        os.mkdir(dbfolder)
        
        self.folder1 = os.path.join(tmpfolder, 'tmp1') + os.sep
        self.folder2 = os.path.join(tmpfolder, 'tmp2') + os.sep
            
        os.mkdir(self.folder1)
        os.mkdir(self.folder2)
            
        os.mkdir(os.path.join(self.folder1, 'folder1'))
        os.mkdir(os.path.join(self.folder1, 'folder2'))
        os.mkdir(os.path.join(self.folder2, 'folder2'))
        os.mkdir(os.path.join(self.folder2, 'folder3'))
        
        self._make_file(os.path.join(self.folder1, 'folder1', 'file1'), 5)
        
        self._make_file(os.path.join(self.folder1, 'folder2', 'file1'), 10)
        self._make_file(os.path.join(self.folder1, 'folder2', 'file2'), 10)
        self._make_file(os.path.join(self.folder2, 'folder2', 'file3'), 10)
        self._make_file(os.path.join(self.folder1, 'file1'), 20)
        
        os.mkdir(os.path.join(self.folder2, 'folder3', 'folder1'))
        os.mkdir(os.path.join(self.folder2, 'folder3', 'folder2'))
        os.mkdir(os.path.join(self.folder2, 'folder3', 'folder3'))
        self._make_file(os.path.join(self.folder2, 'folder3', 'folder1', 'file1'), 5)
        self._make_file(os.path.join(self.folder2, 'folder3', 'folder1', 'file2'), 5)
        self._make_file(os.path.join(self.folder2, 'folder3', 'folder1', 'file3'), 5)
        self._make_file(os.path.join(self.folder2, 'folder3', 'folder2', 'file1'), 5)
        self._make_file(os.path.join(self.folder2, 'folder3', 'folder3', 'file1'), 5)
        
        self.filesystem = FileSystem([(self.folder1, '/sdfs/'), (self.folder2, '/sdfs/')], os.path.join(dbfolder, 'db.db'))
        self.filesystem.fsh.get_time = lambda:1000
        return self.filesystem.rescan()
    
    @defer.inlineCallbacks
    def tearDown(self):
        yield self.filesystem.close()
        shutil.rmtree(self.tmpfolder)
        shutil.rmtree(self.dbfolder)
    
    def testMetadataExists(self):
        self.failIfIdentical(self.filesystem.get_metadata('/sdfs/folder1/'), None, 'No metadata found when there should be.')
    
    def testMetadataDoesNotExist(self):
        self.failUnlessIdentical(self.filesystem.get_metadata('/sdfs/folder10/'), None, 'Metadata found when there should be none.')
    
    def testMultipleCyclesRescan(self):
        import sdfs.fs
        sdfs.fs.COMMIT_COUNTER = 1
        
        def done_scanning(ignored):
            sdfs.fs.COMMIT_COUNTER = 10000
        
        return self.filesystem.rescan().addCallback(done_scanning)
    
    def testListPrefixedSubFolder(self):
        path, listing = self.filesystem.list_dir('/sdfs/').items()[0]
        self.failUnlessEqual(len(listing), 4,
                             "Wrong number of items listed in subfolder")
        self.failUnlessEqual(path, 'sdfs',
                             "Wrong path returned")
        
        for metadata in listing:
            if metadata['type'] == DatabaseType.FILE:
                self.failUnlessIn('size', metadata, 'No size for file')
                self.failUnlessIn('date', metadata, 'No date for file')
                break
        else:
            self.fail("No files found in folder when there should be")
    
    def testListRoot(self):
        path, listing = self.filesystem.list_dir('/').items()[0]
        self.failUnlessEqual(len(listing), 1,
                             "Wrong number of items listed in root")
        self.failUnlessEqual(path, '',
                             "Wrong path returned")
        
        metadata = listing[0]
        self.failUnlessEqual(metadata['type'], DatabaseType.DIRECTORY,
                             "Wrong type detected for folder")
        self.failUnlessEqual(metadata['name'], 'sdfs',
                             "Wrong folder name saved")
        
        self.failUnlessIn('date', metadata, 'Date not in dir listing')
        
    def testListSubSubFolderFromMultipleDisks(self):
        path, listing = self.filesystem.list_dir('/sdfs/folder2/').items()[0]
        self.failUnlessEqual(len(listing), 3,
                             "Wrong number of items listed in multidisk folder")
    
    def testDeleteFolder(self):
        shutil.rmtree(os.path.join(self.folder1, 'folder1'))
        
        def done_scanning(ignored):
            path, listing = self.filesystem.list_dir('/sdfs/').items()[0]
            self.failUnlessEqual(len(listing), 3,
                                "Deleting folder does not work")
        
        return self.filesystem.rescan().addCallback(done_scanning)
    
    def testFileModified(self):
        self.filesystem.fsh.get_time = lambda:2000
        
        def done_scanning(ignored):
            result = self.filesystem.list_dir('/sdfs/')
            for item in result['sdfs']:
                if item['name'] == 'file1':
                    self.failUnlessEqual(item['date'], item['modified'], 'Updated modified date on file')
                elif item['name'] == 'folder3':
                    self.failIfEqual(item['date'], item['modified'], 'Not updated modified date on folder')
                elif item['name'] == 'folder2':
                    self.failUnlessEqual(item['date'], item['modified'], 'Updated modified date on folder')
            
        self._make_file(os.path.join(self.folder2, 'folder3', 'folder3', 'file10'), 5)
        return self.filesystem.rescan().addCallback(done_scanning)
    
    def testDeleteNestedFolder(self):
        shutil.rmtree(os.path.join(self.folder2, 'folder3'))
        
        def done_scanning(ignored):
            path, listing = self.filesystem.list_dir('/sdfs/').items()[0]
            self.failUnlessEqual(len(listing), 3,
                                "Deleting nested folder folder does not work")
        
        return self.filesystem.rescan().addCallback(done_scanning)
    
    def testDeleteFile(self):
        os.remove(os.path.join(self.folder1, 'folder2', 'file1'))
        
        def done_scanning(ignored):
            try:
                path, listing = self.filesystem.list_dir('/sdfs/folder2/').items()[0]
            except:
                self.fail('Exception while listing folder')
            else:
                self.failUnlessEqual(len(listing), 2, "Deleting file does not work")
        
        return self.filesystem.rescan().addCallback(done_scanning)
    
    def testListUnknownPath(self):
        path, listing = self.filesystem.list_dir('unknown/path').items()[0]
        self.failUnlessEqual(path, 'unknown/path', 'Failed to list unknown path')
        self.failUnlessEqual(listing, None, 'Failed to list unknown path')
    
    def testGetUnknownFile(self):
        self.failUnlessIdentical(self.filesystem.get_file('unknown/path'), None, 'Failed to get unknown file')
    
    def testGetFile(self):
        file_endswith = '/tmp1/file1'
        f = self.filesystem.get_file('sdfs/file1')
        if not f.endswith(file_endswith):
            self.fail('Failed to get file, was "%s", expected the following ending: "%s"' % (f, file_endswith))